from typing import Any

import flytekitplugins.omegaconf
import pytest
from flyteidl.core.literals_pb2 import Literal, Scalar
from flytekitplugins.omegaconf.config import OmegaConfTransformerMode
from flytekitplugins.omegaconf.dictconfig_transformer import DictConfigTransformer
from google.protobuf.struct_pb2 import Struct
from omegaconf import MISSING, DictConfig, ListConfig, OmegaConf, ValidationError
from pytest import mark, param

from flytekit import FlyteContext
from flytekit.core.type_engine import TypeEngine
from tests.conftest import ExampleConfig, ExampleNestedConfig
from tests.test_objects import TEST_CFG, MultiTypeEnum, MyConf, MySubConf, SpecialConf


@mark.parametrize(
    ("obj"),
    [
        param(
            DictConfig({}),
        ),
        param(
            DictConfig({"a": "b"}),
        ),
        param(
            DictConfig({"a": 1}),
        ),
        param(
            DictConfig({"a": MISSING}),
        ),
        param(
            DictConfig({"tuple": (1, 2, 3)}),
        ),
        param(
            ListConfig(["a", "b"]),
        ),
        param(
            ListConfig(["a", MISSING]),
        ),
        param(
            TEST_CFG,
        ),
        param(
            OmegaConf.create(ExampleNestedConfig()),
        ),
        param(
            OmegaConf.create(ExampleConfig()),
        ),
        param(
            DictConfig({"foo": MultiTypeEnum.fifo}),
        ),
        param(
            DictConfig({"foo": [MultiTypeEnum.fifo]}),
        ),
        param(DictConfig({"cfgs": [MySubConf(1), MySubConf("a"), "arg"]})),
        param(OmegaConf.structured(SpecialConf)),
    ],
)
def test_cfg_roundtrip(obj: Any) -> None:
    """Test casting DictConfig object to flyte literal and back."""

    ctx = FlyteContext.current_context()
    expected = TypeEngine.to_literal_type(type(obj))
    transformer = TypeEngine.get_transformer(type(obj))

    assert isinstance(
        transformer, flytekitplugins.omegaconf.dictconfig_transformer.DictConfigTransformer
    ) or isinstance(transformer, flytekitplugins.omegaconf.listconfig_transformer.ListConfigTransformer)

    literal = transformer.to_literal(ctx, obj, type(obj), expected)
    reconstructed = transformer.to_python_value(ctx, literal, type(obj))
    assert obj == reconstructed


def test_optional_type() -> None:
    """
    Test serialisation of DictConfigs with various optional entries, whose real types are provided by underlying
    dataclasses.
    """
    optional_obj: DictConfig = OmegaConf.structured(MySubConf())
    optional_obj1: DictConfig = OmegaConf.structured(MyConf(my_attr=MySubConf()))
    optional_obj2: DictConfig = OmegaConf.structured(MyConf())

    ctx = FlyteContext.current_context()
    expected = TypeEngine.to_literal_type(DictConfig)
    transformer = TypeEngine.get_transformer(DictConfig)

    literal = transformer.to_literal(ctx, optional_obj, DictConfig, expected)
    recon = transformer.to_python_value(ctx, literal, DictConfig)
    assert recon == optional_obj

    literal1 = transformer.to_literal(ctx, optional_obj1, DictConfig, expected)
    recon1 = transformer.to_python_value(ctx, literal1, DictConfig)
    assert recon1 == optional_obj1

    literal2 = transformer.to_literal(ctx, optional_obj2, DictConfig, expected)
    recon2 = transformer.to_python_value(ctx, literal2, DictConfig)
    assert recon2 == optional_obj2


def test_plugin_mode() -> None:
    """Test serialisation with different plugin modes configured."""
    obj = OmegaConf.structured(MyConf(my_attr=MySubConf()))

    ctx = FlyteContext.current_context()
    expected = TypeEngine.to_literal_type(DictConfig)

    with flytekitplugins.omegaconf.local_transformer_mode(OmegaConfTransformerMode.DictConfig):
        transformer = DictConfigTransformer()
        literal_slim = transformer.to_literal(ctx, obj, DictConfig, expected)
        reconstructed_slim = transformer.to_python_value(ctx, literal_slim, DictConfig)

    with flytekitplugins.omegaconf.local_transformer_mode(OmegaConfTransformerMode.DataClass):
        literal_full = transformer.to_literal(ctx, obj, DictConfig, expected)
        reconstructed_full = transformer.to_python_value(ctx, literal_full, DictConfig)

    with flytekitplugins.omegaconf.local_transformer_mode(OmegaConfTransformerMode.Auto):
        literal_semi = transformer.to_literal(ctx, obj, DictConfig, expected)
        reconstructed_semi = transformer.to_python_value(ctx, literal_semi, DictConfig)

    assert literal_slim == literal_full == literal_semi
    assert reconstructed_slim == reconstructed_full == reconstructed_semi  # comparison by value should pass

    assert OmegaConf.get_type(reconstructed_slim, "my_attr") == dict
    assert OmegaConf.get_type(reconstructed_semi, "my_attr") == MySubConf
    assert OmegaConf.get_type(reconstructed_full, "my_attr") == MySubConf

    reconstructed_slim.my_attr.my_attr = (1,)  # assign a tuple value to Union[int, str] field
    with pytest.raises(ValidationError):
        reconstructed_semi.my_attr.my_attr = (1,)
    with pytest.raises(ValidationError):
        reconstructed_full.my_attr.my_attr = (1,)


def test_auto_transformer_mode() -> None:
    """Test if auto transformer mode recovers basic information if the specified type cannot be found."""
    obj = OmegaConf.structured(MyConf(my_attr=MySubConf()))

    struct = Struct()
    struct.update(
        {
            "values.my_attr.scalar.union.value.scalar.generic.values.my_attr.scalar.union.value.scalar.primitive.integer": 1,  # noqa: E501
            "values.my_attr.scalar.union.value.scalar.generic.values.my_attr.scalar.union.type.structure.tag": "int",
            "values.my_attr.scalar.union.value.scalar.generic.values.my_attr.scalar.union.type.simple": "INTEGER",
            "values.my_attr.scalar.union.value.scalar.generic.values.list_attr.scalar.generic.values": [],
            "values.my_attr.scalar.union.value.scalar.generic.values.list_attr.scalar.generic.types": [],
            "values.my_attr.scalar.union.value.scalar.generic.types.my_attr": "typing.Union[builtins.int, builtins.str, NoneType]",  # noqa: E501
            "values.my_attr.scalar.union.value.scalar.generic.types.list_attr": "omegaconf.listconfig.ListConfig",
            "values.my_attr.scalar.union.value.scalar.generic.base_dataclass": "tests.test_objects.MySubConf",
            "values.my_attr.scalar.union.type.structure.tag": "OmegaConf DictConfig",
            "values.my_attr.scalar.union.type.simple": "STRUCT",
            "types.my_attr": "typing.Union[omegaconf.dictconfig.DictConfig, NoneType]",
            "base_dataclass": "tests.test_objects.MyConf",
        }
    )
    literal = Literal(scalar=Scalar(generic=struct))

    # construct a literal with an unknown subconfig type
    struct2 = Struct()
    struct2.update(
        {
            "values.my_attr.scalar.union.value.scalar.generic.values.my_attr.scalar.union.value.scalar.primitive.integer": 1,  # noqa: E501
            "values.my_attr.scalar.union.value.scalar.generic.values.my_attr.scalar.union.type.structure.tag": "int",
            "values.my_attr.scalar.union.value.scalar.generic.values.my_attr.scalar.union.type.simple": "INTEGER",
            "values.my_attr.scalar.union.value.scalar.generic.values.list_attr.scalar.generic.values": [],
            "values.my_attr.scalar.union.value.scalar.generic.values.list_attr.scalar.generic.types": [],
            "values.my_attr.scalar.union.value.scalar.generic.types.my_attr": "typing.Union[builtins.int, builtins.str, NoneType]",  # noqa: E501
            "values.my_attr.scalar.union.value.scalar.generic.types.list_attr": "omegaconf.listconfig.ListConfig",
            "values.my_attr.scalar.union.value.scalar.generic.base_dataclass": "tests.test_objects.MyFooConf",
            "values.my_attr.scalar.union.type.structure.tag": "OmegaConf DictConfig",
            "values.my_attr.scalar.union.type.simple": "STRUCT",
            "types.my_attr": "typing.Union[omegaconf.dictconfig.DictConfig, NoneType]",
            "base_dataclass": "tests.test_objects.MyConf",
        }
    )
    literal2 = Literal(scalar=Scalar(generic=struct2))

    ctx = FlyteContext.current_context()
    flytekitplugins.omegaconf.set_transformer_mode(OmegaConfTransformerMode.Auto)
    transformer = DictConfigTransformer()

    reconstructed = transformer.to_python_value(ctx, literal, DictConfig)
    assert obj == reconstructed

    part_reconstructed = transformer.to_python_value(ctx, literal2, DictConfig)
    assert obj == part_reconstructed
    assert OmegaConf.get_type(part_reconstructed, "my_attr") == dict

    part_reconstructed.my_attr.my_attr = (1,)  # assign a tuple value to Union[int, str] field
    with pytest.raises(ValidationError):
        reconstructed.my_attr.my_attr = (1,)



from contextlib import contextmanager

from flytekitplugins.omegaconf.config import OmegaConfTransformerMode
from flytekitplugins.omegaconf.dictconfig_transformer import DictConfigTransformer  # noqa: F401
from flytekitplugins.omegaconf.listconfig_transformer import ListConfigTransformer  # noqa: F401

_TRANSFORMER_MODE = OmegaConfTransformerMode.Auto


def set_transformer_mode(mode: OmegaConfTransformerMode) -> None:
    """Set the global serialization mode for OmegaConf objects."""
    global _TRANSFORMER_MODE
    _TRANSFORMER_MODE = mode


def get_transformer_mode() -> OmegaConfTransformerMode:
    """Get the global serialization mode for OmegaConf objects."""
    return _TRANSFORMER_MODE


@contextmanager
def local_transformer_mode(mode: OmegaConfTransformerMode):
    """Context manager to set a local serialization mode for OmegaConf objects."""
    global _TRANSFORMER_MODE
    previous_mode = _TRANSFORMER_MODE
    set_transformer_mode(mode)
    try:
        yield
    finally:
        set_transformer_mode(previous_mode)


__all__ = ["set_transformer_mode", "get_transformer_mode", "local_transformer_mode", "OmegaConfTransformerMode"]



from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional, Union

from omegaconf import MISSING, OmegaConf


class MultiTypeEnum(str, Enum):
    fifo = "fifo"  # first in first out
    filo = "filo"  # first in last out


@dataclass
class MySubConf:
    my_attr: Optional[Union[int, str]] = 1
    list_attr: List[int] = field(default_factory=list)


@dataclass
class MyConf:
    my_attr: Optional[MySubConf] = None


class SpecialConf(MyConf):
    key: int = 1


TEST_CFG = OmegaConf.create(
    {
        "a": 1,
        "b": 1.0,
        "c": {
            "d": 1,
            "e": MISSING,
            "f": [
                {
                    "g": 2,
                    "h": 1.2,
                },
                {"j": 0.5, "k": "foo", "l": "bar"},
            ],
        },
    }
)



import dataclasses
import typing
from collections import ChainMap

from dataclasses_json import DataClassJsonMixin

from flytekit.loggers import logger
from omegaconf import DictConfig, ListConfig, OmegaConf

NoneType = type(None)


def substitute_types(t: typing.Type) -> typing.Type:
    """
    Provides a substitute type hint to use when selecting transformers for serialisation.

    :param t: Original type
    :return: A corrected typehint
    """
    if hasattr(t, "__origin__"):
        # Only encode generic type and let appropriate transformer handle the rest
        if t.__origin__ in [dict, typing.Dict]:
            t = DictConfig
        elif t.__origin__ in [list, typing.List]:
            t = ListConfig
        else:
            return t.__origin__
    return t


def all_annotations(cls: typing.Type) -> ChainMap:
    """
    Returns a dictionary-like ChainMap that includes annotations for all
    attributes defined in cls or inherited from superclasses.
    """
    return ChainMap(*(c.__annotations__ for c in cls.__mro__ if "__annotations__" in c.__dict__))


def extract_node_type(
    python_val: typing.Union[DictConfig, ListConfig], key: typing.Union[str, int]
) -> typing.Tuple[type, str]:
    """
    Provides typing information about DictConfig nodes

    :param python_val: A DictConfig
    :param key: Key of the node to analyze
    :return:
        - Type - The extracted type
        - str - String representation for (de-)serialisation
    """
    assert isinstance(python_val, DictConfig) or isinstance(
        python_val, ListConfig
    ), "Can only extract type information from omegaconf objects"

    python_val_node_type = OmegaConf.get_type(python_val)
    python_val_annotations = all_annotations(python_val_node_type)

    # Check if type annotations are available
    if hasattr(python_val_node_type, "__annotations__"):
        if key not in python_val_annotations:
            raise ValueError(
                f"Key '{key}' not found in type annotations {python_val_annotations}. "
                "Check your DictConfig object for invalid subtrees not covered by your structured config."
            )

        if typing.get_origin(python_val_annotations[key]) is not None:
            # Abstract types
            origin = typing.get_origin(python_val_annotations[key])
            if getattr(origin, "__name__", None) is not None:
                origin_name = f"{origin.__module__}.{origin.__name__}"
            elif getattr(origin, "_name", None) is not None:
                origin_name = f"{origin.__module__}.{origin._name}"
            else:
                raise ValueError(f"Could not extract name from origin type {origin}")

            # Replace list and dict with omegaconf types
            if origin_name in ["builtins.list", "typing.List"]:
                return ListConfig, "omegaconf.listconfig.ListConfig"
            elif origin_name in ["builtins.dict", "typing.Dict"]:
                return DictConfig, "omegaconf.dictconfig.DictConfig"

            sub_types = []
            sub_type_names = []
            for sub_type in typing.get_args(python_val_annotations[key]):
                if sub_type == NoneType:  # NoneType gets special treatment as no import exists
                    sub_types.append(NoneType)
                    sub_type_names.append("NoneType")
                elif dataclasses.is_dataclass(sub_type) and not issubclass(sub_type, DataClassJsonMixin):
                    # Dataclasses have no matching transformers and get replaced by DictConfig
                    # alternatively, dataclasses can use dataclass_json decorator
                    sub_types.append(DictConfig)
                    sub_type_names.append("omegaconf.dictconfig.DictConfig")
                else:
                    sub_type = substitute_types(sub_type)
                    sub_types.append(sub_type)
                    sub_type_names.append(f"{sub_type.__module__}.{sub_type.__name__}")
            return origin[tuple(sub_types)], f"{origin_name}[{', '.join(sub_type_names)}]"
        elif dataclasses.is_dataclass(python_val_annotations[key]):
            # Dataclasses have no matching transformers and get replaced by DictConfig
            # alternatively, dataclasses can use dataclass_json decorator
            return DictConfig, "omegaconf.dictconfig.DictConfig"
        elif python_val_annotations[key] != typing.Any:
            # Use (cleaned) annotation if it is meaningful
            node_type = substitute_types(python_val_annotations[key])
            type_name = f"{node_type.__module__}.{node_type.__name__}"
            return node_type, type_name

    logger.debug(
        f"Inferring type information directly from runtime object {python_val[key]} for serialisation purposes. "
        "For more stable type resolution and serialisation provide explicit type hints."
    )
    node_type = type(python_val[key])
    type_name = f"{node_type.__module__}.{node_type.__name__}"
    return node_type, type_name
